@@ -941,11 +941,16 @@ message_set_subject (GMimeMessage *message, const char *subject)
 void
 g_mime_message_set_subject (GMimeMessage *message, const char *subject)
 {
+	char *encoded;
+	
 	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	g_return_if_fail (subject != NULL);
 	
 	message_set_subject (message, subject);
-	g_mime_header_set (GMIME_OBJECT (message)->headers, "Subject", message->subject);
+	
+	encoded = g_mime_utils_8bit_header_encode (message->subject);
+	g_mime_header_set (GMIME_OBJECT (message)->headers, "Subject", encoded);
+	g_free (encoded);
 }
 
 
@@ -1265,8 +1270,8 @@ handle_multipart_mixed (GMimeMultipart *multipart, gboolean want_plain, gboolean
 			/* if we haven't yet found a text part or if
                            it is a type we can understand and it is
                            the first of that type, save it */
-			if (!text_part || (!g_strcasecmp (type->subtype, "plain") && 
-					   g_strcasecmp (type->subtype, first_type->subtype) != 0)) {
+			if (!text_part || (!g_strcasecmp (type->subtype, "plain") && (first_type &&
+					   g_strcasecmp (type->subtype, first_type->subtype) != 0))) {
 				*is_html = !g_strcasecmp (type->subtype, "html");
 				text_part = mime_part;
 				first_type = type;
