@@ -57,7 +57,7 @@ struct _GTrie {
 };
 
 
-static __inline__ gunichar
+static inline gunichar
 trie_utf8_getc (const unsigned char **in, size_t inlen)
 {
 	register const unsigned char *inptr = *in;
@@ -126,6 +126,7 @@ g_trie_new (gboolean icase)
 void
 g_trie_free (GTrie *trie)
 {
+	g_ptr_array_free (trie->fail_states, TRUE);
 	memchunk_destroy (trie->match_chunks);
 	memchunk_destroy (trie->state_chunks);
 	g_free (trie);
